#include "cora_fit.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/*  ------------------------ */

static Vector badindex;
static double likezero;

int penalty(Vector *P, Vector *minP, Vector *maxP)
     /*   returns 1 if P<minP or P>maxP
	  which is used as a criterium for a penalty  */
{
  int N, i=0, k=0, flag=0;
  
  N=P->len;

  if (iget_final_result)
    return 0;
  
  /* changed '!=' to '>' to allow for individual lines to have fixed wave/sigma
   */
  if (N > minP->len || N > maxP->len)
    {
      fprintf(stderr, "*  Warning  *: penalty(): different dimensions: "\
	      "N=%d  l=%d  u=%d\n",
	      N, minP->len, maxP->len);
      return(0);
    }

  for(i = 0; i < N; ++i)
    {
      if(P->vec[i] > maxP->vec[i] || P->vec[i] < minP->vec[i]) 
	{
	 badindex.vec[k]=i;
	 ++k;
	 flag = 1;
	}
    }
  badindex.len=k;

  return(flag);
}

/*  ------------------------ 
   the following functions evaluate likelihood
   values called from the Powell routine.
   A function is provided for each set of iterated
   variables. Extra functions are necessary if the
   line distances in multiplets are kept constant
   or/and constant orverall sigmas are to iterated.
  ------------------------ */

/*  Iterate (Lamdba && Sigma)
 *
 *  P(0..nlines-1)                llambda(0..nlines-1)
 *  P(nlines..2*nlines-1)         lsigma(0..nlines-1)  
 */
double func_lik(double * p_vec, int * N)
{
  
  int i, k, flag=0;
  double like;
  
  static Vector ampzero,P;
  static int init=0;
  static Vector minP,maxP;
  
  if (init == 0)
    {
      if (verbose > 0) 
	fprintf(stdout, "Message      : initializing func_lik\n");

      vec_init(&P); 
      vec_alloc(&P,*N);

      vec_init (&minP);
      vec_init (&maxP);
      vec_alloc(&minP,*N);
      vec_alloc(&maxP,*N);

      vec_init(&badindex);
      vec_alloc(&badindex,*N);

      vec_init(&ampzero);
      vec_alloc(&ampzero, nlines);

      k = 0;
      for(i=0; i < nlines; ++i)
	{
	  ampzero.vec[i] = 0.;
	  if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
	    {
	      minP.vec[k]  = lP.vec[i];
	      maxP.vec[k]  = uP.vec[i];
	      ++k;
	    }
	}
      for(i=0;i<nlines;++i)
	{
	  if (fit_sigm[i] == TRUE)
	    {
	      minP.vec[k]  = lP.vec[i+nlines];
	      maxP.vec[k]  = uP.vec[i+nlines];
	      ++k;
	    }
	}

      likezero = likelihood (&ampzero);

/*
    fprintf(stderr,"%f %f %f\n",lsigma.vec[0],mP.vec[1],p_vec[1]);
    for(i=0;i<nlines;++i) {llambda.vec[i]=minP.vec[i];lsigma.vec[i]=minP.vec[nlines+i];}
      blik = callik();
      blik -= likezero;
fprintf(stderr,"%f %f %f\n",lsigma.vec[0],mP.vec[1],p_vec[1]);
    for(i=0;i<nlines;++i) {llambda.vec[i]=p_vec[i];lsigma.vec[i]=p_vec[nlines+i];}
fprintf(stderr,"%f %f %f\n",lsigma.vec[0],mP.vec[1],p_vec[1]);
 */

      init=1;
    }
  
  P.vec = p_vec;
  P.len = *N;

  flag = penalty(&P, &minP, &maxP);

  if(flag > 0) 
      return(LIKE_PENALTY);

  k = 0;
  for(i = 0; i < nlines; ++i)
   {
     if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
       {
	 llambda.vec[i] = P.vec[k];
	 ++k;
       }
   }
  for(i = 0; i < nlines; ++i)
    {
      if (fit_sigm[i] == TRUE)
	{
	  lsigma.vec[i]  = P.vec[k];
	  ++k;
	}
    }

  for(i = 1; i < nlines; ++i)
   {
     if (fit_lock[i] == TRUE)
       llambda.vec[i] = llambda.vec[i-1];
   }


  like = callik();
  like = like - likezero;

/*  if(flag > 0)
    {
     for (i=0;i<badindex.len;++i)
      like = blik+fabs(like)*fabs(P.vec[(int)badindex.vec[i]]-mP.vec[(int)badindex.vec[i]]);
    }
 */

/* note_lik(&P, like, 0); */

  return(like);
}

/*  ------------------------ */

/*  Iterate (Lamdba[0] && Sigma): fixed multiplett
 *
 *  P(0)			llambda(0)
 *  P(1..nlines)		lsigma(0..nlines-1)   
 */
double func_lamclik(double *p_vec, int *N)
{
  int i, k, flag=0;
  double like;
  static int init=0;
  static Vector ampzero,P;
  static Vector minP,maxP;

  if (init == 0)
    {
      if (verbose > 0) 
	fprintf(stdout, "Message      : initializing func_lamclik\n");
      vec_init (&P);
      vec_alloc(&P,*N);
      vec_init (&ampzero);
      vec_init (&minP);
      vec_init (&maxP);
      vec_alloc(&ampzero,nlines);
      vec_alloc(&minP,*N);
      vec_alloc(&maxP,*N);
      
      vec_init(&badindex);
      vec_alloc(&badindex,*N);

      minP.vec[0] = lP.vec[0];
      maxP.vec[0] = uP.vec[0];
      
      k = 1;  
      for(i=0;i<nlines;++i)
	{
	  ampzero.vec[i] = 0.;

	  if (fit_sigm[i] == TRUE)
	    {
	      minP.vec[k]  = lP.vec[i+nlines];
	      maxP.vec[k]  = uP.vec[i+nlines];
	      ++k;
	    }
	}
      likezero = likelihood (&ampzero);
 
      init=1;
    }
  

  P.vec = p_vec;
  P.len = *N;

  flag=penalty(&P,&minP,&maxP);
  if(flag > 0) return(LIKE_PENALTY);

  llambda.vec[0] = P.vec[0];
  for(i=1; i < nlines; ++i)
     if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
         llambda.vec[i] = llambda.vec[i-1] + steplam.vec[i-1];

  k = 1;
  for(i = 0; i < nlines; ++i)
    {
      if (fit_sigm[i] == TRUE)
	{
	  lsigma.vec[i]  = P.vec[k];
	  ++k;
	}
    }
  
  like = callik();
  like = like - likezero;
  
  return(like);
}

/*  ------------------------ */

/*  Iterate (Lamdba[0] && Sigma): fixed multiplett+const sig.
 *
 *  P(0)                        llambda(0)
 *  P(1)     	                lsigma(i); i: first to be iterated
 */
double func_clik(double *p_vec, int *N)
{
  int i, k, flag=0;
  double like;
  static int init=0;
  static Vector ampzero,P;
  static Vector minP,maxP;

  if (init == 0)
    {
      if (verbose > 0)
        fprintf(stdout, "Message      : initializing func_clik\n");
      vec_init (&P);
      vec_alloc(&P,*N);
      vec_init (&ampzero);
      vec_init (&minP);
      vec_init (&maxP);
      vec_alloc(&ampzero,nlines);
      vec_alloc(&minP,*N);
      vec_alloc(&maxP,*N);

      vec_init(&badindex);
      vec_alloc(&badindex,*N);

      minP.vec[0] = lP.vec[0];
      maxP.vec[0] = uP.vec[0];

      k=0;
      while (fit_sigm[k] == FALSE) ++k;
      minP.vec[1]=lP.vec[nlines+k];
      maxP.vec[1]=uP.vec[nlines+k];

      for(i=0;i<nlines;++i)
          ampzero.vec[i] = 0.;
      likezero = likelihood (&ampzero);
      init=1;
    }

  P.vec = p_vec;
  P.len = *N;

  flag=penalty(&P,&minP,&maxP);
  if(flag > 0) return(LIKE_PENALTY);

  llambda.vec[0] =P.vec[0];
  for(i=1; i < nlines; ++i)
    if (fit_wave[i] == TRUE && fit_lock[i] == FALSE) 
      llambda.vec[i] = llambda.vec[i-1] + steplam.vec[i-1];

  for(i = 0; i < nlines; ++i)
      if (fit_sigm[i] == TRUE)
          lsigma.vec[i]  = P.vec[1];

  like = callik();
  like = like - likezero;

  return(like);

}

/*  ------------------------ */

/*  Iterate (Lamdba): fixed sigma
 *
 * P(0..nlines-1)                llambda(0..nlines-1)  
 */
double func_lamlik(double *p_vec, int *N)
{
  int i, k, flag=0;
  double like;
  static int init=0;
  static Vector ampzero,P;
  static Vector minP,maxP;

  if (init == 0)
    {
      if (verbose > 0) 
	fprintf(stdout, "Message      : initializing func_lamlik\n");

      vec_init(&P);vec_alloc(&P,*N);
      vec_init(&ampzero);
      vec_init(&minP);vec_init(&maxP);
      vec_alloc(&ampzero,nlines);
      vec_alloc(&minP,*N);vec_alloc(&maxP,*N);

      vec_init(&badindex);
      vec_alloc(&badindex,*N);

      k = 0;
      for(i=0; i < nlines; ++i)
	{
	  ampzero.vec[i] = 0.;
	  if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
	    {
	      minP.vec[k]    = lP.vec[i];
	      maxP.vec[k]    = uP.vec[i];
	      ++k;
	    }
	}
      likezero = likelihood (&ampzero);
      
      init=1;
    }
  
  P.vec = p_vec;
  P.len = *N;

  flag=penalty(&P, &minP, &maxP);
  if(flag > 0) 
    return(LIKE_PENALTY);

  k = 0;
  for(i=0; i<nlines; ++i)
    {
      if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
	{
	  llambda.vec[i] = P.vec[k];
	  ++k;
	}
    }

  for(i = 1; i < nlines; ++i)
   {
     if (fit_lock[i] == TRUE)
       llambda.vec[i] = llambda.vec[i-1];
   }
  
  like = callik();
  like = like - likezero;
  
  return(like);
}

/*  ------------------------ */

/*  Iterate (Lamdba[0]): fixed sigma, fixed multiplett
 *
 *  P                llambda(0) 
 */
double func_clamlik(double *p_vec, int *N)
{
  int i,flag=0;
  double like;
  static int init=0;
  static Vector ampzero,P;
  static Vector minP,maxP;
  
  if (init == 0)
    {
      if (verbose > 0) 
	fprintf(stdout, "Message      : initializing func_clamlik\n");

      vec_init(&P);
      vec_alloc(&P, *N);
      vec_init(&ampzero);
      vec_alloc(&ampzero, nlines);
      vec_init(&minP);
      vec_init(&maxP);
      vec_alloc(&minP, *N);
      vec_alloc(&maxP, *N);

      vec_init(&badindex);
      vec_alloc(&badindex,*N);

      minP.vec[0]=lP.vec[0];
      maxP.vec[0]=uP.vec[0];

      for(i=0; i<nlines; ++i)
	ampzero.vec[i] = 0.;
      likezero = likelihood (&ampzero);

      init=1;
    }
  
  P.vec = p_vec;
  P.len = *N;
  
  flag = penalty(&P, &minP, &maxP);
  if(flag > 0) 
   return(LIKE_PENALTY);
  
  llambda.vec[0] = P.vec[0];
  for(i=1; i < nlines; ++i)
    if (fit_wave[i] == TRUE && fit_lock[i] == FALSE) 
      llambda.vec[i] = llambda.vec[i-1] + steplam.vec[i-1];
 
  for(i = 1; i < nlines; ++i)
   {
     if (fit_lock[i] == TRUE)
       llambda.vec[i] = llambda.vec[i-1];
   }

  like = callik();
  like = like - likezero;
  
  return(like);
}

/*  ------------------------ */

/*  Iterate (Sigma): fixed lambda
 *
 *  P(0..nlines-1)	lsigma(0..nlines-1)  
 */
double func_siglik(double *p_vec, int *N)
{
  int i, k, flag=0;
  double like;
  static int init=0;
  static Vector ampzero,P;
  static Vector minP,maxP;
  
  if (init == 0)
    {
      if (verbose > 0) 
	fprintf(stdout, "Message      : initializing func_siglik\n");

      vec_init(&P);
      vec_alloc(&P,*N);
      vec_init(&ampzero);
      vec_init(&minP);
      vec_init(&maxP);
      vec_alloc(&minP,*N);
      vec_alloc(&maxP,*N);
      vec_alloc(&ampzero,nlines);

      vec_init(&badindex);
      vec_alloc(&badindex,*N);

      k = 0;
      for(i=0;i<nlines;++i)
	{
	  ampzero.vec[i] = 0.;
	  if (fit_sigm[i] == TRUE)
	    {
	      minP.vec[k] = lP.vec[nlines+i];
	      maxP.vec[k] = uP.vec[nlines+i];
	      ++k;
	    }
	}
      likezero = likelihood (&ampzero);
      
      init=1;
    }
  
  P.vec = p_vec;
  P.len = *N;

  flag  = penalty(&P,&minP,&maxP);
  if(flag > 0) 
    return(LIKE_PENALTY);

  k = 0;
  for(i=0; i < nlines; ++i)
    {
      if (fit_sigm[i] == TRUE)
	{
	  lsigma.vec[i] = P.vec[k];
	  ++k;
	}
    }

  like = callik();
  like = like - likezero;

  return(like);
}

/*  ------------------------ */

/*  Iterate (Lambda,Sigma): const sig.
 *
 *  P(0:nlines-1)      llambda(0:nlines-1)
 *  P(nlines)	       lsigma(i); i first index where sigma >0
 */
double func_sigclik(double *p_vec, int *N)
{

  int i=0, k, flag=0;
  double like;
  static int init=0;
  static Vector ampzero,P;
  static Vector minP,maxP;

  if (init == 0)
    {
      if (verbose > 0)
        fprintf(stdout, "Message      : initializing func_sigclik\n");

      vec_init(&P);
      vec_alloc(&P, *N);
      vec_init(&ampzero);
      vec_alloc(&ampzero, nlines);
      vec_init(&minP);
      vec_init(&maxP);
      vec_alloc(&minP, *N);
      vec_alloc(&maxP, *N);

      vec_init(&badindex);
      vec_alloc(&badindex,*N);

      k=0;
      for(i=0;i<nlines;++i)
        {
          ampzero.vec[i] = 0.;
          if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
            {
              minP.vec[k]  = lP.vec[i];
              maxP.vec[k]  = uP.vec[i];
              ++k;
            }
        }

      i=0;
      while (fit_sigm[i] == FALSE) ++i;
      minP.vec[*N-1]=lP.vec[nlines+i];
      maxP.vec[*N-1]=uP.vec[nlines+i];

      for(i=0; i<nlines; ++i)
        ampzero.vec[i] = 0.;
      likezero = likelihood (&ampzero);

      init=1;
    }

  P.vec = p_vec;
  P.len = *N;

  flag  = penalty(&P,&minP,&maxP);
  if(flag > 0)
    return(LIKE_PENALTY);

  k=0;
  for(i=0; i < nlines; ++i)
    {
      if (fit_wave[i] == TRUE && fit_lock[i] == FALSE)
	{
	  llambda.vec[i]=P.vec[k];
	  ++k;
	}
      if (fit_sigm[i] == TRUE)
          lsigma.vec[i] = P.vec[*N-1];
    }

  for(i = 1; i < nlines; ++i)
   {
     if (fit_lock[i] == TRUE)
       llambda.vec[i] = llambda.vec[i-1];
   }

  like = callik();
  like = like - likezero;

  return(like);

}

/*  ------------------------ */

/*  Iterate (Sigma): fixed lambda, const sig.
 *
 *  P(0)      lsigma(0)
 */
double func_csiglik(double *p_vec, int *N)
{
  int i=0, k, flag=0;
  double like;
  static int init=0;
  static Vector ampzero,P;
  static Vector minP,maxP;

  if (init == 0)
    {
      if (verbose > 0)
        fprintf(stdout, "Message      : initializing func_csiglik\n");

      vec_init(&P);
      vec_alloc(&P, *N);
      vec_init(&ampzero);
      vec_alloc(&ampzero, nlines);
      vec_init(&minP);
      vec_init(&maxP);
      vec_alloc(&minP, *N);
      vec_alloc(&maxP, *N);

      vec_init(&badindex);
      vec_alloc(&badindex,*N);

      k=0;
      while (fit_sigm[k] == FALSE) ++k;
      minP.vec[0]=lP.vec[nlines+k];
      maxP.vec[0]=uP.vec[nlines+k];

      for(i=0; i<nlines; ++i)
        ampzero.vec[i] = 0.;
      likezero = likelihood (&ampzero);

      init=1;
    }

  P.vec = p_vec;
  P.len = *N;

  flag  = penalty(&P,&minP,&maxP);
  if(flag > 0)
    return(LIKE_PENALTY);

  for(i=0; i < nlines; ++i)
    {
      if (fit_sigm[i] == TRUE)
          lsigma.vec[i] = P.vec[0];
    }

  like = callik();
  like = like - likezero;

  return(like);

}
/*  ------------------------ */

/* This function is defined to give the opportunity to calculate
 * likelihood values without changing the fit results globally
 * (e.g. error analysis para_err)
 *
 * P(0..nlines-1)                llambda(0..nlines-1)
 * P(nlines..2*nlines-1)         lsigma(0..nlines-1)   
 */
double func_testlik(Vector *P)
{
  
  int i,flag=0;
  double like;
  
  /*  store variables being changed by callik */
  static double *ycback[MAXLINES];
  static Vector amp,lam,sig;
  
  static Vector ampzero;
  static int init=0;
  
  if (init == 0)
    {
      if (verbose > 0) 
	fprintf(stdout, "Message      : initializing func_testlik\n");

      vec_init(&ampzero);
      vec_alloc(&ampzero,nlines);

      vec_init(&amp);
      vec_init(&lam);
      vec_init(&sig);

      vec_alloc(&amp,nlines);
      vec_alloc(&lam,nlines);
      vec_alloc(&sig,nlines);

      vec_init(&badindex);
      vec_alloc(&badindex,nlines);

      for(i=0;i<nlines;++i)
	{
	  ycback[i] = malloc(count * sizeof(double));
	  if (ycback[i] == NULL)
	    {
	      fprintf (stderr, "* Warning *: Out of memory.\n"); 
	      exit (8);
	    }
	}
      for(i=0;i<nlines;++i)
	{
	  ampzero.vec[i] = 0.;
	}
      likezero = likelihood (&ampzero);
      
      init=1;
    }
  
  flag=penalty(P,&lP,&uP);
  if(flag > 0) 
    return(LIKE_PENALTY);
  
  for(i=0;i<nlines;++i)
    {
      lam.vec[i] = llambda.vec[i];
      sig.vec[i] = lsigma.vec[i];
      llambda.vec[i] = P->vec[i];
      lsigma.vec[i]  = P->vec[nlines+i];
      amp.vec[i] = ampout.vec[i];
      for (flag=0; flag < count; ++flag) 
	ycback[i][flag]=yc[i][flag];
    }

  like = callik();
  like = like - likezero;
  
  /*   restore changed fit results  
   */
  for(i=0; i < nlines; ++i)
    {
      ampout.vec[i] = amp.vec[i];
      for (flag=0; flag < count; ++flag) 
	yc[i][flag] = ycback[i][flag];
      llambda.vec[i] = lam.vec[i];
      lsigma.vec[i]  = sig.vec[i];
    }
  
  return(like);
}

double gcurve(double x, double lam, double sig, double dx)
{

double xx;

xx=exp(-0.5*(x-lam)/sig*(x-lam)/sig)/sig/2.506628275*dx;

return(xx);
}
